import threading

import bpy
from bpy_types import Operator

from . import tasks_queue
from .Tools.helper import *


class FLUENT_OT_BakeMaps(Operator):
    "Bake your PBR maps\nCTRL+SHIFT+ALT for help"
    bl_idname = "fluent.bakepbrmaps"
    bl_label = "Bake PBR maps"
    bl_options = {'REGISTER', 'UNDO'}

    _th = None
    obj = None

    @classmethod
    def poll(cls, context):
        props = bpy.context.scene.FluentShaderProps

        if not props.bake_make_selected_to_active and not bpy.context.active_object:
            cls.poll_message_set("No object selected")
            return False

        if props.bake_make_selected_to_active and len(context.selected_objects) < 2:
            cls.poll_message_set("Baking high to low requires at least two selected objects")
            return False

        if not len(bpy.context.active_object.data.materials):
            cls.poll_message_set("The active object must have a material")
            return False

        if props.bake_in_image and '_' in bpy.context.active_object.name:
            cls.poll_message_set("The object name contains underscore. Please, remove it.")
            return False

        if props.bake_combine_channels:
            missing_bake_channel_name = checkMissingCombineChannel()
            if missing_bake_channel_name is not None:
                cls.poll_message_set(
                    "You selected combine RGB, but map '%s' is not enabled" % missing_bake_channel_name)
                return False

        return True

    def invoke(self, context, event):
        if event.ctrl and event.shift and event.alt:
            bpy.context.window_manager.popup_menu(
                make_oops(['This function bakes what you designed with Fluent Materializer.',
                           'The baking function bakes what is connected to the principled BSDF, which is connected to the material output.',
                           'Be sure the principled BSDF is connected to the material output, then select your object and bake.',
                           'The program will try to save the images in a folder called "Textures" next to the blend file,',
                           'if it can\'t, you have to save your files manualy',
                           'The image name follows this pattern : ObjectName_MAP']), title="About PBR baking",
                icon='INFO')
            return {'FINISHED'}

        self.obj = active_object('GET')
        props = context.scene.FluentShaderProps
        selected_to_active = props.bake_make_selected_to_active
        # Vérification
        if self.obj.type != 'MESH':
            bpy.context.window_manager.popup_menu(
                make_oops(['You try to bake a non mesh thing.']),
                title="Problem detected",
                icon='INFO'
            )

            return {'FINISHED'}
        if not self.obj or len(bpy.context.selected_objects) == 0:
            bpy.context.window_manager.popup_menu(
                make_oops(['No active object.']),
                title="Problem detected",
                icon='INFO'
            )

            return {'FINISHED'}
        elif not len(self.obj.data.materials):
            bpy.context.window_manager.popup_menu(
                make_oops(['No material for the active object.']),
                title="Problem detected",
                icon='INFO'
            )

            return {'FINISHED'}
        elif not len(self.obj.data.uv_layers):
            bpy.context.window_manager.popup_menu(
                make_oops(['No UV map detected.', 'Unwrap the model before to bake.']),
                title="Problem detected",
                icon='INFO'
            )

            return {'FINISHED'}
        elif not selected_to_active and len(bpy.context.selected_objects) >= 2:
            bpy.context.window_manager.popup_menu(
                make_oops(['No selected to active option but more than one object selected.']),
                title="Problem detected",
                icon='INFO'
            )

            return {'FINISHED'}
        # bake high to low
        elif selected_to_active and len(bpy.context.selected_objects) < 2:
            bpy.context.window_manager.popup_menu(
                make_oops(['Select to active option but less than 2 objects selected.']),
                title="Problem detected",
                icon='INFO'
            )

            return {'FINISHED'}

        try:
            bpy.ops.wm.save_mainfile()
        except:
            bpy.context.window_manager.popup_menu(
                make_oops(["Please save your Blender file before baking"]),
                icon="ERROR",
                title="Save your Blender file"
            )
            return {'FINISHED'}

        self._th = threading.Thread(
            target=self.execute_bake_process,
            args=(
                context,
            )
        )

        props.baking_error = ""
        props.is_baking = True
        self._th.start()

        return {'FINISHED'}

    def execute_bake_process(self, context):
        script_file = os.path.dirname(os.path.realpath(__file__)) + '/external_render.py'
        current_blend_file = bpy.data.filepath

        prefs = bpy.context.preferences
        cprefs = prefs.addons['cycles'].preferences
        compute_device_type = cprefs.compute_device_type
        enabled_devices = []
        for device in cprefs.devices:
            if device.use:
                enabled_devices.append(device.id)

        enabled_devices_string = ';'.join(enabled_devices)

        properties = bpy.context.scene.FluentShaderProps
        bake_settings = []
        bake_settings.append(build_string_setting('map_size'))

        bake_image_name = 'None'
        if properties.bake_in_image:
            # Need to print to prevent Blender violation crash ??!! ....
            print('s')
            bake_image_name = properties.bake_in_image.name.split('_')[0]
            properties.bake_in_image_path = properties.bake_in_image.filepath_raw
            # properties.bake_in_image.pack()
            all_maps = find_all_maps_by_image(properties.bake_in_image.name)
            for m in all_maps:
                m.pack()

        bake_settings.append('bake_in_image:'+bake_image_name)
        bake_settings.append(build_string_setting('bake_margin'))
        bake_settings.append(build_string_setting('bake_sample'))
        bake_settings.append(build_string_setting('bake_custom_colorspace'))
        bake_settings.append(build_string_setting('bake_make_selected_to_active'))
        bake_settings.append(build_string_setting('bake_make_selected_to_active_extrusion'))
        bake_settings.append(build_string_setting('bake_use_cage'))

        bake_settings.append(build_string_setting('bake_make_color'))
        bake_settings.append(build_string_setting('bake_make_roughness'))
        bake_settings.append(build_string_setting('bake_make_metallic'))
        bake_settings.append(build_string_setting('bake_make_alpha'))
        bake_settings.append(build_string_setting('bake_make_normal'))
        bake_settings.append(build_string_setting('bake_normal_directx'))
        bake_settings.append(build_string_setting('bake_make_ao'))
        bake_settings.append(build_string_setting('bake_make_emission'))

        bake_settings.append(build_string_setting('bake_combine_channels'))
        bake_settings.append(build_string_setting('bake_red_channel'))
        bake_settings.append(build_string_setting('bake_green_channel'))
        bake_settings.append(build_string_setting('bake_blue_channel'))

        bake_settings.append(build_string_setting('udim_baking'))
        bake_settings.append(build_string_setting('udim_count'))

        bake_settings.append(build_string_setting('bake_image_format'))
        bake_settings_string = ';'.join(bake_settings)

        completed_process = subprocess.run(
            f'{bpy.app.binary_path} {current_blend_file} -b --factory-startup --python-exit-code 1 --python "{script_file}" -- --computedevicetype "{compute_device_type}" --enableddevices "{enabled_devices_string}" --bakesettings "{bake_settings_string}"',
            shell=platform.system() != 'Windows'
        )

        if completed_process.returncode == 0:
            tasks_queue.add_task((after_bake_task, (context, self.obj)))
            return

        tasks_queue.add_task((update_error_baking, (context, "Check the console")))
        tasks_queue.add_task((update_is_baking, (context, False)))


class FLUENT_OT_BakeMask(Operator):
    "Bake edge and cavity mask of the node trees\nCTRL+SHIFT+ALT for help"
    bl_idname = "fluent.bakemask"
    bl_label = "Bake PBR maps"
    bl_options = {'REGISTER', 'UNDO'}

    choice: bpy.props.StringProperty()

    @classmethod
    def poll(cls, context):
        if active_object('GET'):
            return True
        else:
            return False

    def bake_process(self):
        PS = None
        obj_name = None
        mask_nodes = []
        # trouve le principled shader
        node_tree = get_node_tree(self.baked_material)
        nodes = get_nodes(self.baked_material)
        if self.choice == 'ALL':
            mask_nodes = [node for node in nodes if
                          node.type == 'GROUP' and (node.node_tree.name in ['Edges', 'All edges', 'Cavity'])]
        elif self.choice == 'EDGES':
            mask_nodes = [node for node in nodes if
                          node.type == 'GROUP' and node.node_tree.name in ['Edges', 'All edges']]
        elif self.choice == 'CAVITY':
            mask_nodes = [node for node in nodes if node.type == 'GROUP' and 'Cavity' in node.node_tree.name]
        elif self.choice == 'SELECTED':
            mask_nodes = [node for node in nodes if node.select and node.type == 'GROUP' and (
                        node.node_tree.name in ['Edges', 'All edges', 'Cavity'])]
        # On récupère le nom de l'objet sélectionné
        obj = bpy.context.active_object

        if mask_nodes and obj:
            # sauvgarde les paramètres rendu de l'utilisateur
            self.previous_settings['engine'] = bpy.context.scene.render.engine
            self.previous_settings['preview_samples'] = bpy.context.scene.cycles.preview_samples
            self.previous_settings['render_samples'] = bpy.context.scene.cycles.samples
            self.previous_settings['shading_type'] = bpy.context.space_data.shading.type
            # configure le moteur de rendu pour le baking
            bpy.context.scene.render.bake.margin = bpy.context.scene.FluentShaderProps.bake_margin
            bpy.context.scene.cycles.samples = int(bpy.context.scene.FluentShaderProps.bake_sample)
            bpy.context.scene.render.engine = 'CYCLES'
            bpy.context.scene.view_settings.view_transform = 'Standard'

            # Ajout une node emission et la connect à la sortie du matériau
            material_output = get_material_output(nodes)

            try:
                file_path = bpy.data.filepath
                file_name = file_path.split('\\')[-1]
                path = file_path.split(file_name)[0]
                path_texture = path + 'Textures\\'
                if not os.path.exists(path_texture):
                    os.makedirs(path_texture)
                folder_made = True
            except:
                folder_made = False
                self.report({'INFO'}, "The blend file must be saved for automatic image saving.")

            for mask_node in mask_nodes:
                # mute le mask mixer
                for n in mask_node.node_tree.nodes:
                    if n.bl_idname == 'ShaderNodeGroup' and 'Mask Texture Mixer' in n.node_tree.name:
                        n.mute = True

                # ajoute une node image
                img_node = nodes.new(type='ShaderNodeTexImage')

                # créer l'image Mask vide
                mask_img = create_img(self.object, self.size, 'Mask', False, False,
                                      bpy.context.scene.FluentShaderProps.udim_baking,
                                      bpy.context.scene.FluentShaderProps.udim_count)
                if folder_made:
                    mask_img.filepath_raw = path_texture + mask_img.name + '.png'
                mask_img.file_format = 'PNG'

                # Node image utilise l'image Mask
                img_node.image = mask_img

                # Connecte le mask à l'entrée Color de l'émission
                link(self.baked_material, mask_node, 'Mask', material_output, 0)
                # Node image utilise l'image Mask
                img_node.image = mask_img
                # Bake
                # bpy.context.scene.sequencer_colorspace_settings.name = 'sRGB'
                active_object(obj=obj, action='SET', solo=True)
                node_tree.nodes.active = img_node
                bpy.ops.object.bake(type='EMIT')
                # Enregistre l'image
                if folder_made:
                    try:
                        mask_img.save()
                    except:
                        self.report({'INFO'}, "Blender fails to save the files on your system. Save it manually.")
                # Réactive le mask mixer
                # mute le mask mixer
                # for n in mask_node.node_tree.nodes:
                #     if n.bl_idname == 'ShaderNodeGroup' and 'Mask Texture Mixer' in n.node_tree.name:
                #         n.mute = False
                # Ajoute la node bake dans l'arbre
                new_group = import_node_group('Baked')
                original = new_group.node_tree
                single = original.copy()
                new_group.node_tree = single
                new_group.label = 'Baked ' + mask_node.name
                img_node_baked = new_group.node_tree.nodes.get('Image Texture')
                img_node_baked.image = mask_img

                nodes.remove(img_node)

            bpy.context.scene.view_settings.view_transform = 'Filmic'
            bpy.context.scene.render.engine = self.previous_settings['engine']
            bpy.context.scene.cycles.preview_samples = self.previous_settings['preview_samples']
            bpy.context.scene.cycles.samples = self.previous_settings['render_samples']
            bpy.context.space_data.shading.type = self.previous_settings['shading_type']



        else:
            self.report({'ERROR'}, 'No selected object.')

        return {'FINISHED'}

    def invoke(self, context, event):
        if event.ctrl and event.shift and event.alt:
            bpy.context.window_manager.popup_menu(
                make_oops(['This function bake edge and cavity masks from Fluent MC.',
                           'The program searchs and uses the edge and cavity masks from Fluent MC in the node tree.',
                           'The program will try to save the images in a folder called "Textures" next to the blend file if it saved.',
                           'The image name follows this pattern : ObjectName_MaterialName_Mask_NodeName']),
                title="About masks baking", icon='INFO')
            return {'FINISHED'}
        self.previous_settings = {}
        self.size = int(context.scene.FluentShaderProps.map_size)
        self.baked_material = active_object('GET').active_material
        self.object = active_object('GET')

        if not len(active_object('GET').data.uv_layers):
            bpy.context.window_manager.popup_menu(
                make_oops(['No UV map detected.', 'Unwrap the model before to bake.']),
                title="Problem detected",
                icon='INFO'
            )
            return {'FINISHED'}

        self.bake_process()

        return {'FINISHED'}
